<?php

namespace Devsense\PHP\Editor {

    // API interfaces
    require __DIR__ . '\\api.php';

    /**
     * @ignored
     */
    final class Controller implements IIdeController {

        private $extensions = [];

        public function run() {
            ob_start();
            for (;;) {
                $o = self::read_json();
                self::post("return", $this->{$o->Method}($o));
            }
        }

        private static function post($method, $obj) {
            $data = json_encode($obj ?: []) ?: '[]';
            $note = ob_get_contents(); ob_end_clean();
            if ($note != "") { echo "message:" . json_encode($note); }
            echo $method . ':' . $data;
            ob_start();
        }

        private static function read_json() {
            return json_decode(self::consume_json()); // TODO: timeout and exit
        }

        private static function postback($method, $obj) {
            self::post($method, $obj);
            return self::read_json();
        }

        private static function consume_json() {
            $level = 0;
            $escaped = false;
            $quoted = false;
            $result = "";
            
            do {
                $ch = fread(STDIN, 1);
                
                if ($quoted) {
                    if ($escaped) $escaped = false;
                    else if ($ch == '\\') $escaped = true;
                    else if ($ch == $quoted) $quoted = false;
                }
                else {
                    if ($ch == '"' || $ch == "'") $quoted = $ch;
                    else if ($ch == '{' || $ch == '[') $level++;
                    else if ($ch == '}' || $ch == ']') $level--;
                }
                
                $result .= $ch;

            } while ($level != 0 || $quoted);

            //
            return $result;
        }

        #region IIdeController Members

        function output($text) {
            if ($text != "") self::post("message", $text);
        }

        #endregion

        #region API
        
        private function quit($o) { exit; }

        private function load($o) {
            $class = $o->Class;
            if (!array_key_exists($class, $this->extensions)) {
                @include_once $o->FileName;
                if (class_exists($class)) { // && (!method_exists($class,"IsApplicable") || $class::IsApplicable($o->Source))) {
                    $this->extensions[$class] = new $class;
                    return [true];
                }
            }
        }

        private function tooltip($o) {
            $result = null;
            foreach ($this->extensions as $ext) {
            	if ($ext instanceof IPhpEditorCodeSense) {
                    $result[] = $ext->GetToolTip(new NodeInfo($o));
                }
            }
            return $result;
        }

        private function help($o) {
            $result = null;
            foreach ($this->extensions as $ext) {
            	if ($ext instanceof IPhpEditorCodeSense) {
                    $result[] = $ext->GetHelpUrl(new NodeInfo($o));
                }
            }
            return $result;
        }

        private function blocks($o) {
            $result = [];
            foreach ($this->extensions as $ext) {
            	if ($ext instanceof IEditorBlocks) {
                    $result = array_merge($result, $ext->GetBlocks($o->Source));
                }
            }
            return $result;
        }

        private function block_tags($o) {
            $result = [];
            foreach ($this->extensions as $ext) {
            	if ($ext instanceof IEditorBlocks) {
                    $result = array_merge($result, $ext->GetTriggers());
                }
            }
            return $result;
        }

        private function colors($o)
        {
            $spans = null;
            $classifier = $this->extensions[$o->Classifier];
            if ($classifier instanceof IEditorClassifier)
                $spans = $classifier->GetColors();

            return $spans;
        }
        

        #endregion
    }

    //
    (new Controller)->run();
}